# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import unittest
import uuid

from jupyter_client.kernelspec import KernelSpec
from jupyter_client.kernelspec import KernelSpecManager
from jupyter_core.utils import ensure_async, run_sync
from jupyter_server.gateway.gateway_client import GatewayClient
from jupyter_server.gateway.managers import GatewayMappingKernelManager
from jupyter_server.services.kernels.kernelmanager import AsyncMappingKernelManager, ServerKernelManager

from kernels_mixer.kernels import MixingMappingKernelManager
from kernels_mixer.kernelspecs import MixingKernelSpecManager

remote_kernel = "remote-python3"
remote_kernel_manager = KernelSpecManager()
remote_kernel_manager.default_kernel_name=remote_kernel
local_kernel_manager = KernelSpecManager()


def mock_kernel_spec(kernel_name):
    return KernelSpec(name=kernel_name, display_name=kernel_name, language="python")


class MockKernelSpecManager(KernelSpecManager):
    def get_kernel_spec(self, kernel_name, **kwargs):
        return mock_kernel_spec(kernel_name)

    def get_all_specs(self):
        return [self.get_kernel_spec(self.default_kernel_name)]

    def list_kernel_specs(self):
        return {
            "default": self.default_kernel_name,
            "kernelspecs": self.get_all_specs(),
        }


class MockClient():
    async def wait_for_ready(self, *args, **kwargs):
        return


class MockKernelManager(ServerKernelManager):
    def __init__(self, *args, **kwargs):
        super(MockKernelManager, self).__init__(*args, **kwargs)
        self.kernel_id = None
        pass

    def has_kernel(self):
        return self.kernel_id is not None

    def client(self, *args, **kwargs):
        return MockClient()

    async def start_kernel(self, kernel_name=None, kernel_id=None, **kwargs):
        kernel_name = kernel_name or self.parent.kernel_spec_manager.default_kernel_name
        kernel_id = kernel_id or str(uuid.uuid4())
        self.kernel_name = kernel_name
        self.kernel_id = kernel_id
        self.last_activity = datetime.datetime.utcnow()
        self.execution_state = "idle"
        self.ready.set_result(None)
        return kernel_id

    async def shutdown_kernel(self, *args, **kwargs):
        self.kernel_id = None

    async def restart_kernel(self, *args, **kwargs):
        pass

    async def interrupt_kernel(self, *args, **kwargs):
        pass


class MockLocalMappingKernelManager(AsyncMappingKernelManager):
    def __init__(self, *args, **kwargs):
        super(MockLocalMappingKernelManager, self).__init__(*args, **kwargs)
        self.kernel_spec_manager = MockKernelSpecManager()
        self.kernel_manager_class = "kernels_mixer.kernels_test.MockKernelManager"

    async def start_kernel(self, *args, **kwargs):
        km = self.kernel_manager_factory(parent=self)
        kernel_id = await km.start_kernel(*args, **kwargs)
        self._kernels[kernel_id] = km
        return kernel_id

    def _get_changed_ports(self, *args, **kwargs):
        pass


class MockRemoteMappingKernelManager(GatewayMappingKernelManager):
    def __init__(self, *args, **kwargs):
        self.kernel_spec_manager = MockKernelSpecManager()
        self.kernel_spec_manager.default_kernel_name = remote_kernel
        self.kernel_manager_class = "kernels_mixer.kernels_test.MockKernelManager"

    def kernel_model(self, kernel_id):
        return {
            "id": kernel_id,
            "name": self.kernel_spec_manager.default_kernel_name,
            "additional": {
                "foo": "bar",
            },
        }


class MockMixingKernelSpecManager(MixingKernelSpecManager):
    def __init__(self, local_km=None, remote_km=None, **kwargs):
        self.local_manager = local_km or local_kernel_manager
        self.remote_manager= remote_km or remote_kernel_manager
        self._local_kernels = set()
        self._remote_kernels = set()

    def is_remote(self, kernel_name):
        return kernel_name == remote_kernel

class TestKernelModel(unittest.TestCase):
    gwc = GatewayClient.instance()
    gwc.url = gwc.url or "https://example.com"
    gwc.kernelspecs_endpoint = gwc.kernelspecs_endpoint or "/api/kernelspecs"

    local_multikernel_manager = MockLocalMappingKernelManager()
    remote_multikernel_manager =  MockRemoteMappingKernelManager()
    ksm = MockMixingKernelSpecManager(
        local_km=local_multikernel_manager,
        remote_km=remote_multikernel_manager)
    mkm = MixingMappingKernelManager(kernel_spec_manager=ksm)
    mkm.local_manager = local_multikernel_manager
    mkm.remote_manager = remote_multikernel_manager

    async def interrupt_kernel(self, kernel_id):
        await ensure_async(self.mkm.interrupt_kernel(kernel_id))

    def test_local_kernel_model(self):
        local_kernel_name = "python3"
        local_kernel_id = run_sync(self.mkm.start_kernel)(kernel_name=local_kernel_name)
        local_kernel_model = self.mkm.kernel_model(local_kernel_id)
        self.assertEqual(local_kernel_model["id"], local_kernel_id)
        self.assertEqual(local_kernel_model["name"], local_kernel_name)
        self.assertEqual(local_kernel_model["execution_state"], "idle")
        self.assertNotIn("additional", local_kernel_model)
        run_sync(self.interrupt_kernel)(local_kernel_id)
        run_sync(self.mkm.restart_kernel)(local_kernel_id)

    def test_remote_kernel_model(self):
        remote_kernel_id = run_sync(self.mkm.start_kernel)(kernel_name=remote_kernel)
        remote_kernel_model = self.mkm.kernel_model(remote_kernel_id)
        self.assertEqual(remote_kernel_model["id"], remote_kernel_id)
        self.assertEqual(remote_kernel_model["name"], remote_kernel)
        self.assertEqual(remote_kernel_model["additional"]["foo"], "bar")
        self.assertNotIn("execution_state", remote_kernel_model)
        run_sync(self.interrupt_kernel)(remote_kernel_id)
        run_sync(self.mkm.restart_kernel)(remote_kernel_id)


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jupyter_server.gateway.connections import GatewayWebSocketConnection
from jupyter_server.gateway.managers import GatewayKernelManager
from jupyter_server.services.kernels.connection.base import BaseKernelWebsocketConnection
from jupyter_server.services.kernels.connection.channels import ZMQChannelsWebsocketConnection

class DelegatingWebsocketConnection(BaseKernelWebsocketConnection):
    """Implementation of BaseKernelWebsocketConnection that delegates to another connection.

    If the parent KernelManager instance is for a remote kernel (i.e. it is a
    GatewayKernelManager), then the delegate is an instance of GatewayWebSocketConnection.

    Otherwise, it is an instance of ZMQChannelsWebsocketConnection.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        delegate_class = ZMQChannelsWebsocketConnection
        if self.kernel_manager.is_remote:
            delegate_class = GatewayWebSocketConnection
        self.delegate = delegate_class(
            parent=self.kernel_manager.delegate,
            websocket_handler=self.websocket_handler,
            config=self.config)

    async def connect(self):
        return await self.delegate.connect()

    async def disconnect(self):
        return await self.delegate.disconnect()

    def handle_incoming_message(self, msg):
        return self.delegate.handle_incoming_message(msg)

    def handle_outgoing_message(self, stream, msg):
        return self.delegate.handle_outgoing_message(stream, msg)

    # Prepare actually comes from ZMQChannelsWebsocketConnection.
    #
    # It is called by the jupyter_server kernels websocket handler if present, so
    # we provide an implemention of it in case the delegate is an instance of the
    # ZMQChannelWebsocketConnection class.
    async def prepare(self):
        if hasattr(self.delegate, "prepare"):
            return await self.delegate.prepare()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from jupyter_client.manager import in_pending_state
from jupyter_core.utils import ensure_async, run_sync
from jupyter_server.gateway.managers import GatewayMappingKernelManager
from jupyter_server.services.kernels.kernelmanager import AsyncMappingKernelManager
from jupyter_server.services.kernels.kernelmanager import ServerKernelManager

from traitlets import Instance, default

from .kernelspecs import MixingKernelSpecManager

class MixingMappingKernelManager(AsyncMappingKernelManager):

    kernel_spec_manager = Instance(MixingKernelSpecManager)

    @default("kernel_manager_class")
    def _default_kernel_manager_class(self):
        return "kernels_mixer.kernels.MixingKernelManager"
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set up the local kernel management.
        self.local_manager = AsyncMappingKernelManager(
            parent=self.parent,
            log=self.log,
            connection_dir=self.connection_dir,
            kernel_spec_manager=self.kernel_spec_manager.local_manager)

        # Set up the remote kernel management.
        self.remote_manager = GatewayMappingKernelManager(
            parent=self.parent,
            log=self.log,
            connection_dir=self.connection_dir,
            kernel_spec_manager=self.kernel_spec_manager.remote_manager)

    def kernel_model(self, kernel_id):
        self._check_kernel_id(kernel_id)
        kernel = self._kernels[kernel_id]
        return run_sync(kernel.model)()

class MixingKernelManager(ServerKernelManager):
    _kernel_id_map = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def is_remote(self):
        if not self.kernel_name or not self.kernel_id:
            return False
        return self.parent.kernel_spec_manager.is_remote(self.kernel_name)

    @property
    def delegate_kernel_id(self):
        if not self.kernel_id:
            return None
        return MixingKernelManager._kernel_id_map.get(self.kernel_id, None)

    @property
    def delegate_multi_kernel_manager(self):
        if self.is_remote:
            return self.parent.remote_manager
        return self.parent.local_manager

    @property
    def delegate(self):
        if not self.kernel_name or not self.kernel_id:
            return None
        return self.delegate_multi_kernel_manager.get_kernel(self.delegate_kernel_id)

    @property
    def has_kernel(self):
        delegate = self.delegate
        if not delegate:
            return false
        return delegate.has_kernel

    def client(self, *args, **kwargs):
        delegate = self.delegate
        if not delegate:
            return None
        return delegate.client(*args, **kwargs)

    @in_pending_state
    async def start_kernel(self, *args, **kwargs):
        self.kernel_name = kwargs.get("kernel_name", self.kernel_name)
        kernel_id = kwargs.pop("kernel_id", self.kernel_id)
        if kernel_id:
            self.kernel_id = kernel_id
        created_kernel_id = await ensure_async(self.delegate_multi_kernel_manager.start_kernel(
            kernel_name=self.kernel_name, **kwargs))
        MixingKernelManager._kernel_id_map[self.kernel_id] = created_kernel_id

    async def shutdown_kernel(self, *args, **kwargs):
        await ensure_async(self.delegate_multi_kernel_manager.shutdown_kernel(
            self.delegate_kernel_id, *args, **kwargs))
        MixingKernelManager._kernel_id_map.pop(self.kernel_id)

    async def interrupt_kernel(self):
        await ensure_async(self.delegate_multi_kernel_manager.interrupt_kernel(
            self.delegate_kernel_id))

    async def restart_kernel(self, *args, **kwargs):
        await ensure_async(self.delegate_multi_kernel_manager.restart_kernel(
            self.delegate_kernel_id, *args, **kwargs))

    async def model(self):
        delegate_model = await ensure_async(
            self.delegate_multi_kernel_manager.kernel_model(self.delegate_kernel_id))
        model = copy.deepcopy(delegate_model)
        model["id"] = self.kernel_id
        return model



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="kernels-mixer",
  version="0.0.6",
  author="Google, Inc.",
  description="Jupyter server extension that allows mixing local and remote kernels together",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/GoogleCloudPlatform/jupyter-extensions/tree/master/kernels-mixer",
  license="Apache License 2.0",
  packages=setuptools.find_packages(),
  python_requires=">=3.8",
  install_requires=[
    "jupyter_server>=2.7.3",
    "traitlets",
    "google-cloud-jupyter-config",
  ],
)
